/*
 * Copyright (c) 1999-2000 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <ktcommon.h>
#pragma hdrstop

#include <gssapi.h>

#include "ktftp.h"
#include "kclient.h"
#include <base64.h>
#include "ktftp.rh"
#include "ftp.h"
#include "tncodes.h"

//****************************************************************

class FTPSecureKRB4 : public FTPSecure
{
      public:
	FTPSecureKRB4(FTPControlStream *cs);
	~FTPSecureKRB4();

	virtual bool SetSecureLevel(SecureLevel sl);
	virtual SecureLevel GetSecureLevel();
	virtual bool Authenticate(TWindow *parent);
	virtual bool Encode(char * &data, int &count, SecureLevel sl);
	virtual int Decode(char *s, SecureLevel sl);
	virtual bool RemoteKauth(const char *principal, const char *password);
	virtual bool RemoteKlist();
	virtual bool RemoteKdestroy();
	virtual bool IsAuthenticated();
	virtual const char *GetUserName();
	virtual KCPBuf GetKCPBuf();

      protected:
        SecureLevel CurrentSecureLevel;
	FTPControlStream *cs;
	KClient *kc;
	struct krb4_data {
		des_cblock key;
		des_key_schedule schedule;
		char user[ANAME_SZ];
		char instance[INST_SZ];
		char realm[REALM_SZ];
		} app_data;
	struct sockaddr_in localAddr, remoteAddr;

	int MakeAuth(KTEXT adat, char *service, const char *host,
		int checksum);
};

//****************************************************************

class FTPSecureKRB5 : public FTPSecure
{
      public:
	FTPSecureKRB5(FTPControlStream *cs, Kerberos5Context *k5c);
	~FTPSecureKRB5();

	virtual bool SetSecureLevel(SecureLevel sl);
	virtual SecureLevel GetSecureLevel();
	virtual bool Authenticate(TWindow *parent);
	virtual bool Encode(char * &data, int &count, SecureLevel sl);
	virtual int Decode(char *s, SecureLevel sl);
	virtual bool RemoteKauth(const char *principal, const char *password);
	virtual bool RemoteKlist();
	virtual bool RemoteKdestroy();
	virtual bool IsAuthenticated();
	virtual const char *GetUserName();
	virtual KCPBuf GetKCPBuf();

      protected:
	Kerberos5Context *K5C;     
        SecureLevel CurrentSecureLevel;
	FTPControlStream *cs;
	struct sockaddr_in localAddr, remoteAddr;

	struct gss_data {
		gss_ctx_id_t context_hdl;
		char *client_name;
		} GSSData;
      private:
	void SockaddrToGssAddress(const struct sockaddr_in *sa,
		OM_uint32 *addr_type, gss_buffer_desc *gss_addr);
};

//****************************************************************

FTPSecure::FTPSecure(FTPWorker *w)
{
	worker = w;
}

FTPSecure::~FTPSecure()
{
}

bool
FTPSecure::SetSecureLevel(SecureLevel sl)
{
	return (sl == slNone);
}

FTPSecure::SecureLevel
FTPSecure::GetSecureLevel()
{
	return slNone;
}

bool
FTPSecure::Authenticate(TWindow *)
{
	worker->CMDPrintf(IDS_NOAUTH);
	return true;
}


bool
FTPSecure::Encode(char * &Data, int &Count, SecureLevel /*sl*/)
{
	int NewCount;
	char *NewData;

	NewCount = strlen(Data) + 3;
	if ((NewData = (char *)malloc(NewCount)) == NULL) {
		worker->CMDPrintf(IDS_OOFM);
		return false;
		}

	strlcpy(NewData, Data, NewCount);
	strlcat(NewData, "\r\n", NewCount);
	free(Data);
	Data = NewData;
	Count = NewCount - 1;
	return true;
}

int
FTPSecure::Decode(char *, SecureLevel)
{
	return -1;			// Can't decode unless coding is on!
}

KCPBuf
FTPSecure::GetKCPBuf()
{
	return KCPBuf(0, 0, 0, 0);
}

#pragma argsused
bool
FTPSecure::RemoteKauth(const char *principal, const char *password)
{
	return false;
}

bool
FTPSecure::RemoteKlist()
{
	return false;
}

bool
FTPSecure::RemoteKdestroy()
{
	return false;
}

bool
FTPSecure::IsAuthenticated()
{
	return false;
}

const char *
FTPSecure::GetUserName()
{
	return "";
}

//****************************************************************

FTPSecureKRB4::FTPSecureKRB4(FTPControlStream *_cs)
	: FTPSecure(_cs->GetWorker())
{
	cs = _cs;
	kc = cs->GetKClient();
	CurrentSecureLevel = slNone;
}

FTPSecureKRB4::~FTPSecureKRB4()
{
}

bool
FTPSecureKRB4::Authenticate(TWindow *)
{
	int ret;
	char *p;
	int len;
	KTEXT_ST adat;
	MSG_DAT msg_data;
	int checksum, rcs;

	checksum = 0xdeadbeef;	//getpid();

	strlcpy(app_data.realm,
		krb_realmofhost(cs->GetHostName()),
		sizeof(app_data.realm));

	ret = MakeAuth(&adat, "ftp", cs->GetHostName(), checksum);

	if (ret == KDC_PR_UNKNOWN)
		ret = MakeAuth(&adat, "rcmd", cs->GetHostName(), checksum);

	if (ret == KFAILURE) {
		worker->CMDPrintf(IDS_OPC);
		return false;
		}	

	if (ret) {
		worker->CMDPrintf(IDS_UTGT, krb_get_err_text(ret));
		worker->RetryConnection(krb_get_err_text(ret));
		return false;
		}

	kc->GetUserName(app_data.user);

	localAddr = *cs->GetLocalAddr();
	remoteAddr = *cs->GetRemoteAddr();

	if (krb_get_config_bool("nat_in_use")) {
		struct in_addr natAddr;

		if (wkrb_get_our_ip_for_realm(app_data.realm, &natAddr) != KSUCCESS)
			worker->CMDPrintf(IDS_WCGNIPA, app_data.realm);
		else {
			if (natAddr.s_addr != localAddr.sin_addr.s_addr) {
				worker->CMDPrintf(IDS_UNIPA, inet_ntoa(natAddr));
				localAddr.sin_addr = natAddr;

				// This not the best place to do this, but it
				// is here we know that (probably) NAT is in
				// use!

				worker->CMDPrintf(IDS_TONPM);
				worker->PassiveMode = true;
				}
			}
		}

	worker->CMDPrintf(IDS_LAIS, inet_ntoa(localAddr.sin_addr));
	worker->CMDPrintf(IDS_RAIS, inet_ntoa(remoteAddr.sin_addr));

	kc->GetSessionKey((KClientKey *)app_data.key);
	des_key_sched(&app_data.key, app_data.schedule);

	if (base64_encode(adat.dat, adat.length, &p) < 0) {
		worker->CMDPrintf(IDS_OOFM);
		return false;
		}
	ret = cs->Command("ADAT %s", p);
	free(p);

	if (ret != FTP_COMPLETE) {
		worker->CMDPrintf(IDS_SDAAD);
		worker->RetryConnection(cs->ReplyString);
		return false;
		}

	p = strstr(cs->ReplyString, "ADAT=");
	if (!p) {
		worker->CMDPrintf(IDS_SDSSR);
		return false;
		}
	p += 5;
	len = base64_decode(p, adat.dat);
	if (len < 0) {
		worker->CMDPrintf(IDS_FTDB64);
		return false;
		}
	adat.length = len;
	ret = krb_rd_safe(adat.dat, adat.length, (des_cblock *)app_data.key, 
		&remoteAddr, &localAddr, &msg_data);
	if (ret) {
		worker->CMDPrintf(IDS_ERRFS, krb_get_err_text(ret));
		return false;
		}
	krb_get_int(msg_data.app_data, (u_int32_t *)&rcs, 4, 0);
	if (rcs - checksum != 1) {
		worker->CMDPrintf(IDS_BADCK);
		return false;
		}
	SetSecureLevel(slSafe);
	return true;
}

int
FTPSecureKRB4::MakeAuth(KTEXT adat, 
	char *service, const char *host, int checksum)
{
	int result;
	char principal[255];
	unsigned long adatLen = sizeof (KTEXT_ST);

	// Get Kerberos ticket
	principal[0] = '\0';
	strlcat(principal, service, sizeof(principal));
	strlcat(principal, ".", sizeof(principal));
	strlcat(principal, krb_get_phost(host), sizeof(principal));
	strlcat(principal,"@", sizeof(principal));
	strlcat(principal, krb_realmofhost(host), sizeof(principal));

	cs->GetWorker()->CMDPrintf(IDS_TRYP, principal);

	result = kc->GetTicketForService(principal, adat, &adatLen,
		checksum, true);

	return result;
}

bool
FTPSecureKRB4::SetSecureLevel(SecureLevel sl)
{
	switch (sl) {
		case slNone:
		case slSafe:
		case slPrivate:
			CurrentSecureLevel = sl;
			return true;
		}
	return false;
}

FTPSecure::SecureLevel
FTPSecureKRB4::GetSecureLevel()
{
	return CurrentSecureLevel;
}

bool
FTPSecureKRB4::Encode(char * &Data, int &Count, SecureLevel sl)
{
	char *NewData;
	char *EncodedData;
	char *cmd;
	int NewCount;

	if (sl == slDefault)
		sl = CurrentSecureLevel;

	if (sl == slNone)
		return FTPSecure::Encode(Data, Count, slNone);

	if ((NewData = (char *)malloc(Count + 31)) == NULL) {
		worker->CMDPrintf(IDS_OOFM);
		return false;
		}

	switch (sl) {
		case slSafe:
			NewCount = krb_mk_safe(Data, NewData, Count,
				&app_data.key, &localAddr, &remoteAddr);
			cmd = "MIC ";
			break;
		case slPrivate:
			NewCount = krb_mk_priv(Data, NewData, Count,
				app_data.schedule, &app_data.key,
				&localAddr, &remoteAddr);
			cmd = "ENC ";
			break;
		default:
			free(NewData);
			worker->CMDPrintf("Internal error: Bad protection level\n");
			return false;
		}

	if (base64_encode(NewData, NewCount, &EncodedData) < 0) {
		worker->CMDPrintf(IDS_OOFM);
		free(NewData);
		return false;
		}
	free(NewData);

//	worker->CMDPrintf("Sending: MIC %s\n", EncodedData);
	NewCount = strlen(EncodedData) + 7;
	
	if ((NewData = (char *)malloc(NewCount)) == NULL) {
		worker->CMDPrintf(IDS_OOFM);
		free(EncodedData);
		return false;
		}

	strlcpy(NewData, cmd, NewCount);
	strlcat(NewData, EncodedData, NewCount);
	strlcat(NewData, "\r\n", NewCount);

	free(Data);
	Data = NewData;
	Count = NewCount - 1;
	return true;
}

int
FTPSecureKRB4::Decode(char *s, SecureLevel sl)
{
	char *WorkBuf;
	int WorkLen;
	int KRBErr;
	int Code;
	MSG_DAT Msg;
    
	if (sl == slDefault)
		sl = CurrentSecureLevel;

	if (sl == slNone)
		return FTPSecure::Decode(s, slNone);

	if ((WorkBuf = (char *)malloc(strlen(s))) == NULL) {
		worker->CMDPrintf(IDS_OOFM);
		return -1;
		}

	if ((WorkLen = base64_decode(s + 4, WorkBuf)) < 0) {
		worker->CMDPrintf(IDS_FTDB64);
		free(WorkBuf);
		return -1;
		}

	switch (sl) {
		case slSafe:
			KRBErr = krb_rd_safe(WorkBuf, WorkLen, &app_data.key,
				&remoteAddr, &localAddr, &Msg);
			break;
		case slPrivate:
			KRBErr = krb_rd_priv(WorkBuf, WorkLen,
				app_data.schedule, &app_data.key,
				&remoteAddr, &localAddr, &Msg);
			break;
		default:
			free(WorkBuf);
			worker->CMDPrintf("Internal error, bad protection level\n");
			return -1;
		}

	if (KRBErr) {
		worker->CMDPrintf(IDS_CDD);
		return -1;
		}

	MoveMemory(WorkBuf, Msg.app_data, Msg.app_length);
	WorkLen = Msg.app_length;

	if (WorkLen > 2) {
		if ((WorkBuf[WorkLen-2] == '\r') &&
		    (WorkBuf[WorkLen-1] == '\n'))
			WorkLen -= 2;
		}

	WorkBuf[WorkLen] = '\0';

	if (WorkBuf[3] == '-')
		Code = 0;
	else
		sscanf(WorkBuf, "%d", &Code);

//	worker->CMDPrintf("Decoded data: %s\n", WorkBuf);
	MoveMemory(s, WorkBuf, WorkLen+1);
	free(WorkBuf);
	return Code;
}

KCPBuf
FTPSecureKRB4::GetKCPBuf()
{
	return kc->GetKCPBuf();
}

bool
FTPSecureKRB4::RemoteKauth(const char *principal, const char *password)
{
	int ret;
	des_cblock key;
	des_key_schedule schedule;
	KTEXT_ST tkt, tktcopy;
	char *p;
	char passwd[100];
	int tmp;
	char buf[INST_SZ+ANAME_SZ];

	if (!principal) {
		kc->GetUserName(buf);
		principal = buf;
		}
	ret = cs->Command(slPrivate, "SITE KAUTH %s", principal);
	if (ret != FTP_CONTINUE)
		return false;

	p = strstr(cs->ReplyString, "T=");
	if (!p) {
		worker->CMDPrintf(IDS_BADRPLY);
		return false;
		}

	p += 2;
	tmp = base64_decode(p, &tkt.dat);
	if (tmp < 0) {
		worker->CMDPrintf(IDS_FTDB64);
		return false;
		}
	tkt.length = tmp;
	tktcopy.length = tkt.length;
    
	p = strstr(cs->ReplyString, "P=");
	if (!p) {
		worker->CMDPrintf(IDS_BADRPLY);
		return false;
		}
	principal = p + 2;
	for(; *p && *p != ' ' && *p != '\r' && *p != '\n'; p++);
	*p = 0;
    
	if (!password) {
		char prompt[sizeof(buf)+128];

		strlcpy(prompt, worker->ResString(IDS_PRIPASS),
			sizeof(prompt));
		strlcat(prompt, principal, sizeof(prompt));
		strlcat(prompt, ": ", sizeof(prompt));

		passwd[0] = 0;
		if (!worker->GetUserPrompt(FTPWorker::ptOther, prompt, passwd,
			sizeof(passwd), false))
			return false;
		password = passwd;
		}

	des_string_to_key ((char *)password, &key);

	des_key_sched(&key, schedule);
    
	des_pcbc_encrypt((des_cblock*)tkt.dat, (des_cblock*)tktcopy.dat,
		tkt.length, schedule, &key, DES_DECRYPT);

	if (strcmp ((char*)tktcopy.dat + 8, 
		KRB_TICKET_GRANTING_TICKET) != 0) {
	        afs_string_to_key((char *)password, krb_realmofhost(cs->GetHostName()),
			&key);
		des_key_sched (&key, schedule);
		des_pcbc_encrypt((des_cblock*)tkt.dat,
			(des_cblock*)tktcopy.dat, tkt.length,
			schedule, &key, DES_DECRYPT);
		}
	memset(key, 0, sizeof(key));
	memset(schedule, 0, sizeof(schedule));
	memset(passwd, 0, sizeof(passwd));
	if (base64_encode(tktcopy.dat, tktcopy.length, &p) < 0) {
		worker->CMDPrintf(IDS_OOFM);
		return false;
		}
	memset(tktcopy.dat, 0, tktcopy.length);
	ret = cs->Command(slPrivate, "SITE KAUTH %s %s", principal, p);
	free(p);
	return ret == FTP_COMPLETE;
}

bool
FTPSecureKRB4::RemoteKlist()
{
	return cs->Command("SITE KLIST") == FTP_COMPLETE;
}

bool
FTPSecureKRB4::RemoteKdestroy()
{
	return cs->Command("SITE KDESTROY") == FTP_COMPLETE;
}

bool
FTPSecureKRB4::IsAuthenticated()
{
	return CurrentSecureLevel != slNone;
}

const char *
FTPSecureKRB4::GetUserName()
{
	return app_data.user;
}

//****************************************************************

FTPSecureKRB5::FTPSecureKRB5(FTPControlStream *_cs, Kerberos5Context *k5c)
	: FTPSecure(_cs->GetWorker())
{
	cs = _cs;
	K5C = k5c;
	CurrentSecureLevel = slNone;

	GSSData.context_hdl = GSS_C_NO_CONTEXT;
}

FTPSecureKRB5::~FTPSecureKRB5()
{
}

void
FTPSecureKRB5::SockaddrToGssAddress(const struct sockaddr_in *sa,
	OM_uint32 *addr_type,
	gss_buffer_desc *gss_addr)
{
	gss_addr->length = 4;
	gss_addr->value  = (struct sockaddr_in *)&sa->sin_addr;
	*addr_type       = GSS_C_AF_INET;
}

bool
FTPSecureKRB5::Authenticate(TWindow *parent)
{
	OM_uint32 maj_stat, min_stat;
	gss_buffer_desc name;
	gss_name_t target_name;
	gss_buffer_desc input, output_token;
	int context_established = 0;
	char *p;
	gss_channel_bindings_t bindings;
	    
	name.length = asprintf((char**)&name.value,
		"ftp@%s", cs->GetHostName());
		
	maj_stat = gss_import_name(&min_stat,
		&name, GSS_C_NT_HOSTBASED_SERVICE, &target_name);
	if (GSS_ERROR(maj_stat)) {
		OM_uint32 new_stat;
		OM_uint32 msg_ctx = 0;
		gss_buffer_desc status_string;
	    
		gss_display_status(&new_stat,
			min_stat,
			GSS_C_MECH_CODE,
			GSS_C_NO_OID,
			&msg_ctx,
			&status_string);
		worker->CMDPrintf("Error importing name %s: %s\n", 
			(char *)name.value,
			(char *)status_string.value);
		gss_release_buffer(&new_stat, &status_string);
		worker->RetryConnection("GSSAPI error");
		return false;
		}
	free(name.value);

	if (!K5C->MakeRequestLogin(parent, target_name)) {
		worker->RetryConnection("Can't login");
		return false;
		}

	input.length = 0;
	input.value = NULL;

	bindings = (gss_channel_bindings_t)malloc(sizeof(*bindings));

	SockaddrToGssAddress(cs->GetLocalAddr(),
		&bindings->initiator_addrtype,
		&bindings->initiator_address);
	SockaddrToGssAddress(cs->GetRemoteAddr(),
		&bindings->acceptor_addrtype,
		&bindings->acceptor_address);

	bindings->application_data.length = 0;
	bindings->application_data.value = NULL;

	while (!context_established) {
		maj_stat = gss_init_sec_context(&min_stat,
			GSS_C_NO_CREDENTIAL,
			&GSSData.context_hdl,
			target_name,
			GSS_C_NO_OID,
			GSS_C_MUTUAL_FLAG | GSS_C_SEQUENCE_FLAG
				| GSS_C_DELEG_FLAG,
			0,
			bindings,
			&input,
			NULL,
			&output_token,
			NULL,
			NULL);
		if (GSS_ERROR(maj_stat)) {
			OM_uint32 new_stat;
			OM_uint32 msg_ctx = 0;
			gss_buffer_desc status_string;
			
			gss_display_status(&new_stat,
				min_stat,
				GSS_C_MECH_CODE,
				GSS_C_NO_OID,
				&msg_ctx,
				&status_string);
			worker->CMDPrintf("Error initializing security context: %s\n", 
				(char*)status_string.value);
			gss_release_buffer(&new_stat, &status_string);
			worker->RetryConnection("GSSAPI error");
			return false;
			}

		gss_release_buffer(&min_stat, &input);
		if (output_token.length != 0) {
			base64_encode(output_token.value, output_token.length, &p);
			gss_release_buffer(&min_stat, &output_token);
			cs->Command("ADAT %s", p);
			free(p);
			}
		if (GSS_ERROR(maj_stat)) {
			if (GSSData.context_hdl != GSS_C_NO_CONTEXT)
				gss_delete_sec_context (&min_stat,
					&GSSData.context_hdl,
					GSS_C_NO_BUFFER);
				break;
			}
		if (maj_stat & GSS_S_CONTINUE_NEEDED) {
			p = strstr(cs->ReplyString, "ADAT=");
			if (p == NULL) {
				worker->CMDPrintf("Error: expected ADAT in reply.\n");
				return false;
				}
			else {
				p += 5;
				input.value = malloc(strlen(p));
				input.length = base64_decode(p, input.value);
				}
			}
		else {
			if (cs->LastCode != 235) {
				worker->CMDPrintf("Unrecognized response code: %d\n", cs->LastCode);
				return false;
				}
			context_established = 1;
			}
		}


	SetSecureLevel(slSafe);
	return true;
}

bool
FTPSecureKRB5::SetSecureLevel(SecureLevel sl)
{
	switch (sl) {
		case slNone:
		case slSafe:
		case slPrivate:
			CurrentSecureLevel = sl;
			return true;
		}
	return false;
}

FTPSecure::SecureLevel
FTPSecureKRB5::GetSecureLevel()
{
	return CurrentSecureLevel;
}

bool
FTPSecureKRB5::Encode(char * &Data, int &Count, SecureLevel sl)
{
	char *NewData;
	int NewCount;
	char *EncodedData;
	char *cmd;
	OM_uint32 min_stat;
	gss_buffer_desc input, output;
	int conf_state;

	if (sl == slDefault)
		sl = CurrentSecureLevel;

	if (sl == slNone)
		return FTPSecure::Encode(Data, Count, slNone);

	input.length = Count;
	input.value = Data;
	gss_wrap (&min_stat,
		GSSData.context_hdl,
		sl == slPrivate,
		GSS_C_QOP_DEFAULT,
		&input,
		&conf_state,
		&output);

	NewData = (char *)output.value;
	NewCount = output.length;

	switch (sl) {
		case slSafe:
			cmd = "MIC ";
			break;
		case slPrivate:
			cmd = "ENC ";
			break;
		default:
			free(NewData);
			worker->CMDPrintf("Internal error: Bad protection level\n");
			return false;
		}

	if (base64_encode(NewData, NewCount, &EncodedData) < 0) {
		worker->CMDPrintf(IDS_OOFM);
		free(NewData);
		return false;
		}
	free(NewData);

//	worker->CMDPrintf("Sending: MIC %s\n", EncodedData);
	NewCount = strlen(EncodedData) + 7;
	
	if ((NewData = (char *)malloc(NewCount)) == NULL) {
		worker->CMDPrintf(IDS_OOFM);
		free(EncodedData);
		return false;
		}

	strlcpy(NewData, cmd, NewCount);
	strlcat(NewData, EncodedData, NewCount);
	strlcat(NewData, "\r\n", NewCount);

	free(Data);
	Data = NewData;
	Count = NewCount - 1;

	return true;
}

int
FTPSecureKRB5::Decode(char *s, SecureLevel sl)
{
	char *WorkBuf;
	int WorkLen;
	int Code;
	OM_uint32 maj_stat, min_stat;
	gss_buffer_desc input, output;
	gss_qop_t qop_state;
	int conf_state;
    
	if (sl == slDefault)
		sl = CurrentSecureLevel;

	if (sl == slNone)
		return FTPSecure::Decode(s, slNone);

	if ((WorkBuf = (char *)malloc(strlen(s))) == NULL) {
		worker->CMDPrintf(IDS_OOFM);
		return -1;
		}

	if ((WorkLen = base64_decode(s + 4, WorkBuf)) < 0) {
		worker->CMDPrintf(IDS_FTDB64);
		free(WorkBuf);
		return -1;
		}

	input.length = WorkLen;
	input.value = WorkBuf;
	maj_stat = gss_unwrap (&min_stat,
		GSSData.context_hdl,
		&input,
		&output,
		&conf_state,
		&qop_state);
	if (GSS_ERROR(maj_stat)) {
		worker->CMDPrintf(IDS_CDD);
		return -1;
		}

	MoveMemory(WorkBuf, output.value, output.length);
	WorkLen = output.length;

	if (WorkLen > 2) {
		if ((WorkBuf[WorkLen-2] == '\r') &&
		    (WorkBuf[WorkLen-1] == '\n'))
			WorkLen -= 2;
		}

	WorkBuf[WorkLen] = '\0';

	if (WorkBuf[3] == '-')
		Code = 0;
	else
		sscanf(WorkBuf, "%d", &Code);

//	worker->CMDPrintf("Decoded data: %s\n", WorkBuf);
	MoveMemory(s, WorkBuf, WorkLen+1);
	free(WorkBuf);
	return Code;
}

KCPBuf
FTPSecureKRB5::GetKCPBuf()
{
	return K5C->GetKCPBuf();
}

bool
FTPSecureKRB5::RemoteKauth(const char *principal, const char *password)
{
	int ret;
	des_cblock key;
	des_key_schedule schedule;
	KTEXT_ST tkt, tktcopy;
	char *p;
	char passwd[100];
	int tmp;

	if (!principal)
		principal = K5C->GetUser();

	worker->CMDPrintf("Using <%s>\n", principal);

	ret = cs->Command(slPrivate, "SITE KAUTH %s", principal);
	if (ret != FTP_CONTINUE)
		return false;

	p = strstr(cs->ReplyString, "T=");
	if (!p) {
		worker->CMDPrintf(IDS_BADRPLY);
		return false;
		}

	p += 2;
	tmp = base64_decode(p, &tkt.dat);
	if (tmp < 0) {
		worker->CMDPrintf(IDS_FTDB64);
		return false;
		}
	tkt.length = tmp;
	tktcopy.length = tkt.length;
    
	p = strstr(cs->ReplyString, "P=");
	if (!p) {
		worker->CMDPrintf(IDS_BADRPLY);
		return false;
		}
	principal = p + 2;
	for(; *p && *p != ' ' && *p != '\r' && *p != '\n'; p++);
	*p = 0;
    
	if (!password) {
		char prompt[1024];

		strlcpy(prompt, worker->ResString(IDS_PRIPASS),
			sizeof(prompt));
		strlcat(prompt, principal, sizeof(prompt));
		strlcat(prompt, ": ", sizeof(prompt));

		passwd[0] = 0;
		if (!worker->GetUserPrompt(FTPWorker::ptOther, prompt, passwd,
			sizeof(passwd), false))
			return false;
		password = passwd;
		}

	des_string_to_key ((char *)password, &key);

	des_key_sched(&key, schedule);
    
	des_pcbc_encrypt((des_cblock*)tkt.dat, (des_cblock*)tktcopy.dat,
		tkt.length, schedule, &key, DES_DECRYPT);

	if (strcmp ((char*)tktcopy.dat + 8, 
		KRB_TICKET_GRANTING_TICKET) != 0) {
	        afs_string_to_key((char *)password, krb_realmofhost(cs->GetHostName()),
			&key);
		des_key_sched (&key, schedule);
		des_pcbc_encrypt((des_cblock*)tkt.dat,
			(des_cblock*)tktcopy.dat, tkt.length,
			schedule, &key, DES_DECRYPT);
		}
	memset(key, 0, sizeof(key));
	memset(schedule, 0, sizeof(schedule));
	memset(passwd, 0, sizeof(passwd));
	if (base64_encode(tktcopy.dat, tktcopy.length, &p) < 0) {
		worker->CMDPrintf(IDS_OOFM);
		return false;
		}
	memset(tktcopy.dat, 0, tktcopy.length);
	ret = cs->Command(slPrivate, "SITE KAUTH %s %s", principal, p);
	free(p);
	return ret == FTP_COMPLETE;
}

bool
FTPSecureKRB5::RemoteKlist()
{
	return cs->Command("SITE KLIST") == FTP_COMPLETE;
}

bool
FTPSecureKRB5::RemoteKdestroy()
{
	return cs->Command("SITE KDESTROY") == FTP_COMPLETE;
}

bool
FTPSecureKRB5::IsAuthenticated()
{
	return CurrentSecureLevel != slNone;
}

const char *
FTPSecureKRB5::GetUserName()
{
	return K5C->GetUser();
}

//****************************************************************

static bool
TryKerberos4(FTPControlStream *cs)
{
	FTPWorker *w;
	int ret;

	if ((cs->GetKClient()) == NULL)
		return false;		// Not supported

	w = cs->GetWorker();	// Needed for printing

	ret = cs->Command("AUTH KERBEROS_V4");
	if (ret != FTP_CONTINUE) {
		if (cs->LastCode == 504)
			w->CMDPrintf(IDS_KV4NS);
		else if (cs->LastCode == 534)
			w->CMDPrintf(IDS_KV4RJ);
		else if (ret == FTP_ERROR)
			w->CMDPrintf(IDS_NOSEC);
		return false;
		}
	return true;		// Go ahead and try....
}

static bool
TryGSSAPI(FTPControlStream *cs)
{
	FTPWorker *w;
	int ret;

	if ((cs->GetKClient()) == NULL)
		return false;		// Not supported

	w = cs->GetWorker();	// Needed for printing

	ret = cs->Command("AUTH GSSAPI");
	if (ret != FTP_CONTINUE) {
		if (cs->LastCode == 504)
			w->CMDPrintf(IDS_KV4NS);
		else if (cs->LastCode == 534)
			w->CMDPrintf(IDS_KV4RJ);
		else if (ret == FTP_ERROR)
			w->CMDPrintf(IDS_NOSEC);
		return false;
		}
	return true;		// Go ahead and try....
}

//****************************************************************

FTPSecure *
GetSecurityHandler(FTPControlStream *cs, Kerberos5Context *k5c)
{
	if (k5c)
		if (TryGSSAPI(cs))
			return new FTPSecureKRB5(cs, k5c);

	if (TryKerberos4(cs))
		return new FTPSecureKRB4(cs);

	return new FTPSecure(cs->GetWorker());	// Default, no security
}

