/*
 * Copyright (c) 2000 Thomas Nystrom & Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "krb5_locl.h"
#include <Windows.h>

void
msg(char *text, int error)
{
    char *buf;

    asprintf (&buf, "%s\nAn error of type: %d", text, error);

    MessageBox(GetActiveWindow(),
	       buf ? buf : "Out of memory!",
	       "Kerberos Message",
	       MB_OK|MB_APPLMODAL);
    free (buf);
}

int
krb_registry_read(const char *key, const char *subkey,
	char *value, size_t value_size)
{
	char k[128];
	HKEY Handler, Root;
	DWORD DataType;
	DWORD DataSize = value_size;
	
	if (strcasecmp(key, "Manager") == 0)
		Root = HKEY_CURRENT_USER;
	else
		Root = HKEY_LOCAL_MACHINE;

	strlcpy(k, "SOFTWARE\\Kerberos\\", sizeof(k));
	strlcat(k, key, sizeof(k));

      again:
	if (RegOpenKeyEx(Root, k, 0, KEY_READ, &Handler) !=
		ERROR_SUCCESS)
		goto failed;

	if (RegQueryValueEx(Handler,
		subkey,
		NULL,
		&DataType,
		(BYTE *)value,
		&DataSize) != ERROR_SUCCESS) {
		RegCloseKey(Handler);
		goto failed;
		}

	RegCloseKey(Handler);
	if ((DataType == REG_DWORD) && (value_size >= sizeof(int))) {
		if (*(int *)value)
			value[0] = '1';
		else
			value[0] = '0';
		DataType = REG_SZ;
		DataSize = 1;
		}

	if (DataType != REG_SZ)
		return 0;
	
	value[DataSize] = 0;

	return 1;

      failed:
	if (Root != HKEY_CURRENT_USER)
		return 0;
	strlcpy(k, ".Default\\SOFTWARE\\Kerberos\\", sizeof(k));
	strlcat(k, key, sizeof(k));
	Root = HKEY_USERS;
	goto again;
}

int
krb_registry_enumkey(const char *key, int idx,
	char *subkey, size_t subkeylen)
{
	char k[128];
	HKEY Handler, Root;
	DWORD SubKeyLen = subkeylen - 1;
	FILETIME ft;
	
	if (strcasecmp(key, "Manager") == 0)
		Root = HKEY_CURRENT_USER;
	else
		Root = HKEY_LOCAL_MACHINE;

	strlcpy(k, "SOFTWARE\\Kerberos\\", sizeof(k));
	strlcat(k, key, sizeof(k));

	if (RegOpenKeyEx(Root, k, 0, KEY_READ, &Handler) !=
		ERROR_SUCCESS)
		return 0;		// Not found!

	if (RegEnumKeyEx(Handler, idx, subkey, &SubKeyLen,
		NULL, NULL, NULL, &ft) != ERROR_SUCCESS) {
		RegCloseKey(Handler);
		return 0;
		}
	RegCloseKey(Handler);

	subkey[SubKeyLen] = 0;

	return 1;
}

int
krb_registry_enumstring(const char *key, int idx,
	char *label, size_t labellen,
	char *value, size_t valuelen)
{
	char k[128];
	HKEY Handler, Root;
	DWORD type;
	DWORD LabelLen = labellen - 1;
	DWORD ValueLen = valuelen - 1;
	
	if (strcasecmp(key, "Manager") == 0)
		Root = HKEY_CURRENT_USER;
	else
		Root = HKEY_LOCAL_MACHINE;

	strlcpy(k, "SOFTWARE\\Kerberos\\", sizeof(k));
	strlcat(k, key, sizeof(k));

	if (RegOpenKeyEx(Root, k, 0, KEY_READ, &Handler) !=
		ERROR_SUCCESS)
		return 0;		// Not found!

	if (RegEnumValue(Handler, idx, label, &LabelLen, NULL, &type,
		(LPBYTE)value, &ValueLen) != ERROR_SUCCESS) {
		RegCloseKey(Handler);
		return 0;
		}
	RegCloseKey(Handler);

	label[LabelLen] = 0;
	value[ValueLen] = 0;

	return 1;
}

void
StartTicketManager(void)
{
	char TicketManager[256], *n;
	STARTUPINFO StartInfo;
	PROCESS_INFORMATION ProcessInfo;

	ZeroMemory(&StartInfo, sizeof(StartInfo));
	StartInfo.cb = sizeof(StartInfo);

	if (FindWindow("KtelnetTicketManager", NULL) != NULL) {
		// Already have ticket manager running!
		return;
		}

	n = NULL;
	if (krb_registry_read("Manager", "AutoStart",
		TicketManager, sizeof(TicketManager))) {
		if (TicketManager[0] == '0')
			return;		// Don't start TicketManager
		}

	if (!krb_registry_read("Manager", "Exefile",
		TicketManager, sizeof(TicketManager))) {

		if (GetModuleFileName(GetModuleHandle("KRB.DLL"),
			TicketManager, sizeof(TicketManager))) {
			if ((n = strrchr(TicketManager, '\\')) == NULL)
				n = strrchr(TicketManager, ':');
			}

		if (n == NULL)
			n = TicketManager - 1;

		strlcpy(n + 1, "ktelnet -m",
			sizeof(TicketManager) - ((n + 1) - TicketManager));
		}
	CreateProcess(0, TicketManager, 0, 0, FALSE, 0, 0, 0,
		&StartInfo, &ProcessInfo);
}

BOOL WINAPI
DllEntryPoint (HANDLE hInst, 
	 ULONG reason,
	 LPVOID lpReserved)
{
	WORD wVersionRequested; 
	WSADATA wsaData; 
	int err; 

	switch (reason) {
		case DLL_PROCESS_ATTACH:
			wVersionRequested = MAKEWORD(1, 1); 
			err = WSAStartup(wVersionRequested, &wsaData); 
			if (err != 0) {
				/* Tell the user that we couldn't find a */
				/* useable winsock.dll.     */ 
				msg("Cannot find winsock.dll", err);
				return FALSE;
				}
#if 0
			if (newTktMem(0) != KSUCCESS) {
				/* Tell the user that we couldn't alloc */
				/* shared memory. */ 
				msg("Cannot allocate shared ticket memory",
					GetLastError());
				return FALSE;
				}
			if (krb_get_config_string("win32_debug")) {
				krb_debug = 1;
				krb_dns_debug = 1;
				}
#endif
			break;

		case DLL_PROCESS_DETACH:
			/* should this really be done here? */
//			freeTktMem(0);
			WSACleanup();
			break;
		}
	return TRUE;
}

krb5_error_code
PostUpdateMessage(krb5_error_code cod)
{
	HWND hWnd;

	hWnd = FindWindow("KtelnetTicketManager", NULL);
	if (hWnd == NULL)
		hWnd = HWND_BROADCAST;
	PostMessage(hWnd, RegisterWindowMessage("KtelnetTicketUpdate"), 0, 0);
	return cod;
}

#if 0
void
dumpit(int id, FILE *fp, krb5_config_section *cs)
{
	while (cs) {
		fprintf(fp, "%*.*s%s = ", id, id, "", cs->name);
		switch (cs->type) {
			case krb5_config_string:
				fprintf(fp, "%s\n", cs->u.string);
				break;
			case krb5_config_list:
				fprintf(fp, "{\n");
				dumpit(id+4, fp, cs->u.list);
				fprintf(fp, "%*.*s}\n", id, id, "");
				break;
			default:
				fprintf(fp, "<<UNKNOWN TYPE>>\n");
			}
		cs = cs->next;
		}
}

void
dump_config(krb5_config_section *cs)
{
	FILE *fp;
	
//	MessageBox(0, "Dumping", "DEBUG", MB_OK);
	fp = fopen("\\config.dmp", "w");
	if (fp == NULL)
		return;
	
	dumpit(1, fp, cs);
	fclose(fp);
}
#endif
